import enum
import re

def escape(string):
	result = ''
	for char in string:
		if char == '\'':
			result += '\\27'
		elif char == '\\':
			result += '\\5C'
		else:
			result += char
	return result

class Syntax:
	oid: str
	desc: str

	def __init__(self, max_len=None):
		self.max_len = max_len
		if max_len is None:
			self.ref = self.oid
		else:
			self.ref = self.oid + '{' + str(max_len) + '}'

	@classmethod
	def get_first_component_oid(cls):
		return cls.oid

	@classmethod
	def to_definition(cls):
		return f"( {cls.oid} DESC '{escape(cls.desc)}' )"

	def decode(self, schema, raw_value):
		'''Decode LDAP-specific encoding of a value to a native value

		:param raw_value: LDAP-specific encoding of the value
		:type raw_value: bytes

		:returns: native value (depends on syntax), None if raw_value is invalid
		:rtype: any or None'''
		return None

	def encode(self, schema, value):
		'''Encode native value to its LDAP-specific encoding

		:param value: native value (depends on syntax)
		:type value: any

		:returns: LDAP-specific encoding of the value
		:rtype: bytes'''
		raise NotImplementedError()

class MatchingRule:
	def __init__(self, oid, name, syntax, **kwargs):
		self.oid = oid
		self.name = name
		self.names = [name]
		self.syntax = syntax
		for key, value in kwargs.items():
			setattr(self, key, value)

	def to_definition(self):
		return f"( {self.oid} NAME '{escape(self.name)}' SYNTAX {self.syntax.ref} )"

	def __repr__(self):
		return f'<ldapserver.schema.MatchingRule {self.oid}>'

	def match_equal(self, schema, attribute_value, assertion_value):
		return None

	def match_approx(self, schema, attribute_value, assertion_value):
		return self.match_equal(schema, attribute_value, assertion_value)

	def match_less(self, schema, attribute_value, assertion_value):
		return None

	def match_greater_or_equal(self, schema, attribute_value, assertion_value):
		return None

	def match_substr(self, schema, attribute_value, inital_substring, any_substrings, final_substring):
		return None

class AttributeTypeUsage(enum.Enum):
	# pylint: disable=invalid-name
	# user
	userApplications = enum.auto()
	# directory operational
	directoryOperation = enum.auto()
	# DSA-shared operational
	distributedOperation = enum.auto()
	# DSA-specific operational
	dSAOperation = enum.auto()

class AttributeType:
	# pylint: disable=too-many-instance-attributes,too-many-arguments,too-many-branches,too-many-statements
	def __init__(self, oid, name=None, desc=None, obsolete=None, sup=None,
	             equality=None, ordering=None, substr=None, syntax=None,
               single_value=None, collective=None, no_user_modification=None,
               usage=None):
		if sup is None and syntax is None:
			raise ValueError('Either SUP or, syntax=syntax.must, be specified')
		tokens = ['(', oid]
		if name is not None:
			tokens += ['NAME', "'"+escape(name)+"'"] # name is actually a list
		if desc is not None:
			tokens += ['DESC', "'"+escape(desc)+"'"]
		if obsolete is not None:
			tokens += ['OBSOLETE', obsolete]
		if sup is not None:
			tokens += ['SUP', sup.oid]
		if equality is not None:
			tokens += ['EQUALITY', equality.oid]
		if ordering is not None:
			tokens += ['ORDERING', ordering.oid]
		if substr is not None:
			tokens += ['SUBSTR', substr.oid]
		if syntax is not None:
			tokens += ['SYNTAX', syntax.ref]
		if single_value is not None:
			tokens += ['SINGLE-VALUE']
		if collective is not None:
			tokens += ['COLLECTIVE']
		if no_user_modification is not None:
			tokens += ['NO-USER-MODIFICATION']
		if usage is not None:
			tokens += ['USAGE', usage.name]
		tokens += [')']
		self.schema_encoding = ' '.join(tokens)
		self.oid = oid
		self.name = name
		self.names = [name] if name is not None else []
		self.inherited_names = set(self.names)
		self.obsolete = obsolete or False
		self.sup = sup
		if self.sup is not None:
			self.inherited_names |= self.sup.inherited_names
		self.equality = equality
		if self.equality is None and self.sup is not None:
			self.equality = self.sup.equality
		self.ordering = ordering
		if self.ordering is None and self.sup is not None:
			self.ordering = self.sup.ordering
		self.substr = substr
		if self.substr is None and self.sup is not None:
			self.substr = self.sup.substr
		self.syntax = syntax
		if self.syntax is None and self.sup is not None:
			self.syntax = self.sup.syntax
		self.single_value = single_value or False
		self.collective = collective or False
		self.no_user_modification = no_user_modification or False
		self.usage = usage or AttributeTypeUsage.userApplications

	def to_definition(self):
		return self.schema_encoding

	def get_first_component_oid(self):
		return self.oid

	def __repr__(self):
		return f'<ldapserver.schema.AttributeType {self.oid}>'

class ObjectClassKind(enum.Enum):
	ABSTRACT = enum.auto()
	STRUCTURAL = enum.auto()
	AUXILIARY = enum.auto()

class ObjectClass:
	# pylint: disable=too-many-arguments
	def __init__(self, oid, name=None, desc=None, obsolete=None, sup=None,
	             kind=None, must=None, may=None):
		tokens = ['(', oid]
		if name is not None:
			tokens += ['NAME', "'"+escape(name)+"'"] # name is actually a list
		if desc is not None:
			tokens += ['DESC', "'"+escape(desc)+"'"]
		if obsolete is not None:
			tokens += ['OBSOLETE', obsolete]
		if sup is not None:
			tokens += ['SUP', sup.name]
		if kind is not None:
			tokens += [kind.name]
		if must and len(must) == 1:
			tokens += ['MUST', must[0].name]
		elif must and len(must) > 1:
			tokens += ['MUST', '(']
			for index, attr in enumerate(must):
				if index > 0:
					tokens += ['$']
				tokens += [attr.name]
			tokens += [')']
		if may and len(may) == 1:
			tokens += ['MAY', may[0].name]
		elif may and len(may) > 1:
			tokens += ['MAY', '(']
			for index, attr in enumerate(may):
				if index > 0:
					tokens += ['$']
				tokens += [attr.name]
			tokens += [')']
		tokens += [')']
		self.schema_encoding = ' '.join(tokens)
		self.oid = oid
		self.name = name
		self.names = [name]
		self.desc = desc
		self.obsolete = obsolete or False
		self.sup = sup
		self.kind = kind or ObjectClassKind.STRUCTURAL
		self.must = must or []
		self.may = may or []

	def to_definition(self):
		return self.schema_encoding

	def get_first_component_oid(self):
		return self.oid

	def __repr__(self):
		return f'<ldapserver.schema.ObjectClass {self.schema_encoding}>'

DOTTED_DECIMAL_RE = re.compile(r'[0-9]+(\.[0-9]+)*')

def normalize_oid(oid):
	return oid.lower().strip()

class Schema:
	'''Collection of LDAP syntaxes, matching rules, attribute types and object
	classes forming an LDAP schema.'''
	def __init__(self, object_classes=None, attribute_types=None, matching_rules=None, syntaxes=None):
		self.syntaxes = []
		self.matching_rules = []
		self.attribute_types = []
		self.user_attribute_types = []
		self.object_classes = []
		self.__attribute_type_by_oid = {}
		self.__attribute_type_and_subtypes_by_oid = {}
		self.__oid_names = {}
		for syntax in syntaxes:
			self.register_syntax(syntax)
		for matching_rule in matching_rules:
			self.register_matching_rule(matching_rule)
		for attribute_type in attribute_types:
			self.register_attribute_type(attribute_type)
		for object_class in object_classes:
			self.register_object_class(object_class)

	def extend(self, *schemas, object_classes=None, attribute_types=None, matching_rules=None, syntaxes=None):
		object_classes = self.object_classes + (object_classes or [])
		attribute_types = self.attribute_types + (attribute_types or [])
		matching_rules = self.matching_rules + (matching_rules or [])
		syntaxes = self.syntaxes + (syntaxes or [])
		for schema in schemas:
			object_classes += schema.object_classes
			attribute_types += schema.attribute_types
			matching_rules += schema.matching_rules
			syntaxes += schema.syntaxes
		return Schema(object_classes, attribute_types, matching_rules, syntaxes)

	def register_syntax(self, syntax):
		'''Add syntax to schema

		:param syntax: Syntax (subclass, not instance!) to add
		:type syntax: Syntax subclass'''
		if syntax in self.syntaxes:
			return
		self.register_oid(syntax.oid)
		self.syntaxes.append(syntax)

	def register_matching_rule(self, matching_rule):
		'''Add matching rule and the referenced syntax to schema

		:param matching_rule: Matching rule to add
		:type matching_rule: MatchingRule'''
		if matching_rule in self.matching_rules:
			return
		self.register_syntax(matching_rule.syntax)
		self.register_oid(matching_rule.oid, *matching_rule.names)
		self.matching_rules.append(matching_rule)

	def register_attribute_type(self, attribute_type):
		'''Add attribute type and all referenced matching rules and syntaxes to schema

		:param attribute_type: Attribute type to add
		:type attribute_type: AttributeType'''
		if attribute_type in self.attribute_types:
			return
		self.register_syntax(attribute_type.syntax)
		if attribute_type.equality:
			self.register_matching_rule(attribute_type.equality)
		if attribute_type.ordering:
			self.register_matching_rule(attribute_type.ordering)
		if attribute_type.substr:
			self.register_matching_rule(attribute_type.substr)
		self.register_oid(attribute_type.oid, *attribute_type.names)
		for oid in [attribute_type.oid] + attribute_type.names:
			self.__attribute_type_by_oid[normalize_oid(oid)] = attribute_type
			self.__attribute_type_and_subtypes_by_oid.setdefault(normalize_oid(oid), [])
			self.__attribute_type_and_subtypes_by_oid[normalize_oid(oid)].append(attribute_type)
		for oid in attribute_type.inherited_names:
			self.__attribute_type_and_subtypes_by_oid.setdefault(normalize_oid(oid), [])
			self.__attribute_type_and_subtypes_by_oid[normalize_oid(oid)].append(attribute_type)
		self.attribute_types.append(attribute_type)
		if attribute_type.usage == AttributeTypeUsage.userApplications:
			self.user_attribute_types.append(attribute_type)

	def register_object_class(self, object_class):
		'''Add object class and all referenced attribute types, matching rules and syntaxes to schema

		:param object_class: Object class to add
		:type object_class: ObjectClass'''
		if object_class in self.object_classes:
			return
		for attribute_type in object_class.may + object_class.must:
			self.register_attribute_type(attribute_type)
		self.register_oid(object_class.oid, *object_class.names)
		self.object_classes.append(object_class)

	def get_attribute_type(self, oid):
		'''Get attribute type by its OID

		:param oid: Numeric OID or short descriptive name of attribute type
		:type oid: str
		:return: Attribute type identified by OID
		:rtype: AttributeType
		:raises KeyError: if attribute type is not found'''
		attribute_type = self.__attribute_type_by_oid.get(normalize_oid(oid))
		if attribute_type is None:
			raise KeyError(f'Attribute "{oid}" not in schema')
		return attribute_type

	def get_attribute_type_and_subtypes(self, oid):
		'''Get attribute type by its OID with all subtypes (if any)

		:param oid: Numeric OID or short descriptive name of attribute type
		:type oid: str
		:return: List containing the attribute type and its subtypes or empty list
		         if attribute type is not found
		:rtype: List[AttributeType]'''
		return self.__attribute_type_and_subtypes_by_oid.get(normalize_oid(oid), [])

	def register_oid(self, numeric_oid, *names):
		'''Register numeric OID and optionally short descriptive names with the schema

		Both the numeric OID and all names must be unique within the schema.

		:param numeric_oid: Numeric OID
		:type numeric_oid: str
		:param names: Short descriptive names for OID
		:type names: str'''
		for name in names:
			if self.__oid_names.get(normalize_oid(name), numeric_oid) != numeric_oid:
				raise Exception(f'OID short descriptive name "{name}" is already used in schema')
		for name in names:
			self.__oid_names[normalize_oid(name)] = numeric_oid

	def get_numeric_oid(self, oid):
		'''Return numeric OID for a given OID short descriptive name

		If `oid` is a numeric OID it is returned normalized.

		:param oid: OID short descriptive name or numeric OID
		:type oid: str
		:return: Numeric OID in dotted-decimal form or None if `oid` is not
		         recognized and is not a numeric OID.
		:rtype: str or None'''
		oid = normalize_oid(oid)
		if DOTTED_DECIMAL_RE.fullmatch(oid):
			return oid
		return self.__oid_names.get(oid)
